# SPDX-License-Identifier: MIT


# Imports.
from avro.core import config


# Functions.
def count_vowels(text: str) -> int:
    """Count number of occurrences of vowels in a given string.

    Parameters:
    -----------

    text: str
        The text to count vowels from.

    Returns:
    --------

    int
        The number of vowels in the given text.
    """

    return sum(1 for i in text if i.lower() in config.AVRO_VOWELS)


def count_consonants(text: str) -> int:
    """Count number of occurrences of consonants in a given string.

    Parameters:
    -----------

    text: str
        The text to count consonants from.

    Returns:
    --------

    int
        The number of consonants in the given text.
    """

    return sum(1 for i in text if i.lower() in config.AVRO_CONSONANTS)



# SPDX-License-Identifier: MIT


"""A modern Pythonic implementation of the popular Bangla typing software Avro Phonetic.

Overview
--------

avro.py provides a fully fledged, batteries-included text parser which can
parse, reverse and even convert English Roman script into its phonetic
equivalent (unicode) of Bengali.

Features
--------

1. Simple-to-use syntax with functional approach.
2. Easily convert/parse/reverse English Roman script to Bengali phonetic and vice versa.
3. Also asynchronous and `async`/`await` sugarcoated syntax compliant.
4. Synchronous alternative functions available for applications requiring minimal overhead.
5. Batteries included. Just write your text and let it do the rest.

Examples
--------

Please refer to the following link to view the examples:
https://github.com/hitblast/avro.py/tree/main/examples

"""

# Import local modules.
from .main import *

# Version information.
__version_info__ = (2024, 12, 9)
__version__ = ".".join(map(str, __version_info__))



# SPDX-License-Identifier: MIT


# Imports.
import contextlib
from functools import lru_cache
from typing import Any, Optional

from . import config, validate

# Setup pattern variables for matching.
PATTERNS = config.DICT["avro"]["patterns"]
NON_RULE_PATTERNS = [p for p in PATTERNS if "rules" not in p]
RULE_PATTERNS = [p for p in PATTERNS if "rules" in p]


# ---

# The following functions provide the functionality for the main "avro" package.
# These include rule / non-rule based pattern matching, remapping, reversing and more.
# Please do not modify these functions if you do not have the intentions to alter the core functionality of the module.

# ---


@lru_cache(maxsize=128)
def find_in_remap(text: str, *, reversed: bool = False) -> tuple[str, bool]:
    """Finds and returns the remapped value for a given text.

    Parameters:
    -----------

    text: str
        The text to be remapped.

    reversed: bool
        Whether to operate in reverse mode.

    Returns:
    --------

    tuple[str, bool]
        A tuple of two elements:
        1. The remapped text.
        2. Whether manual intervention is required.
    """

    previous_text = text

    for key, value in config.AVRO_EXCEPTIONS.items():
        if reversed:
            text = (
                text.replace(key, value)
                if key.lower() in text.lower()
                else text
            )
        else:
            text = (
                text.replace(value, key)
                if (value := value.lower()) in text.lower()
                else text
            )

    manual_required = any(
        word == previous_word
        for word, previous_word in zip(text.split(), previous_text.split())
    )

    return text, manual_required


def match_patterns(
    fixed_text: str, cur: int = 0, rule: bool = False, reversed: bool = False
) -> dict[str, Any]:
    """Matches given text at cursor position with rule / non rule patterns.

    Parameters:
    -----------

    fixed_text: str
        The text to be matched.

    cur: int
        The cursor position.

    rule: bool
        Whether to match with rules.

    reversed: bool
        Whether to operate in reverse mode.

    Returns:
    --------

    dict[str, Any]
        A dictionary containing the following keys:
        1. matched: bool
        2. found: str
        3. replaced: str
        4. reversed: str
        5. rules: dict[str, Any]
    """

    rule_type = NON_RULE_PATTERNS if not rule else RULE_PATTERNS
    pattern = exact_find_in_pattern(fixed_text, reversed, cur, rule_type)

    if pattern:
        p = pattern[0]

        return {
            "matched": True,
            "found": p.get("find"),
            "replaced": p.get("replace"),
            "reversed": reverse_with_rules(cur, fixed_text, p.get("reverse"))
            if not rule
            else None,
            "rules": p.get("rules") if rule else None,
        }

    return {
        "matched": False,
        "found": None,
        "replaced": fixed_text[cur],
        "rules": None if rule else None,
    }


def exact_find_in_pattern(
    fixed_text: str, reversed: bool, cur: int = 0, patterns: Any = PATTERNS
) -> list[dict[str, Any]]:
    """Returns pattern items that match given text, cursor position and pattern.

    Parameters:
    -----------

    fixed_text: str
        The text to be matched.

    reversed: bool
        Whether to operate in reverse mode.

    cur: int = 0
        The cursor position.

    patterns: Any = PATTERNS
        The patterns to be matched.

    Returns:
    --------

    list[dict[str, Any]]
        A list of dictionaries containing the matched patterns.
    """

    if reversed:
        return [
            x
            for x in patterns
            if (cur + len(x["replace"]) <= len(fixed_text))
            and x["replace"] == fixed_text[cur : (cur + len(x["replace"]))]
        ]

    return [
        x
        for x in patterns
        if x.get("find", None)
        and (cur + len(x["find"]) <= len(fixed_text))
        and x["find"] == fixed_text[cur : (cur + len(x["find"]))]
    ]


def reverse_with_rules(
    cursor: int, fixed_text: str, text_reversed: str
) -> str:
    """Enhances the word with rules for reverse-parsing.

    Parameters:
    -----------

    cursor: int
        The cursor position.

    fixed_text: str
        The fixed text.

    text_reversed: str
        The reversed text.

    Returns:
    --------

    str
        The enhanced word.
    """

    added_suffix = ""

    if not (
        fixed_text[cursor] in config.AVRO_KAR
        or fixed_text[cursor] in config.AVRO_SHORBORNO
        or fixed_text[cursor] in config.AVRO_IGNORE
        or len(fixed_text) == cursor + 1
    ):
        added_suffix = "o"

    with contextlib.suppress(IndexError):
        if (fixed_text[cursor + 1] in config.AVRO_KAR) or (
            fixed_text[cursor + 2] in config.AVRO_KAR and not cursor == 0
        ):
            added_suffix = ""

    return text_reversed if not text_reversed else text_reversed + added_suffix


def process_rules(
    rules: dict[str, Any], fixed_text: str, cur: int = 0, cur_end: int = 1
) -> Optional[str]:
    """Process rules matched in pattern and returns suitable replacement.

    If any rule's condition is satisfied, output the rules "replace", else output None.

    Parameters:
    -----------

    rules: dict[str, Any]
        The rules to be processed.

    fixed_text: str
        The fixed text.

    cur: int = 0
        The cursor position.

    cur_end: int = 1
        The end cursor position.

    Returns:
    --------

    Optional[str]
        The replaced text if any rule's condition is satisfied, else None.
    """

    replaced = ""

    # Iterate through rules.
    for rule in rules:
        matched = False

        for match in rule["matches"]:
            matched = process_match(match, fixed_text, cur, cur_end)

            if not matched:
                break

        if matched:
            replaced = rule["replace"]
            break

    return replaced if matched else None


def process_match(match: Any, fixed_text: str, cur: int, cur_end: int) -> bool:
    """Processes a single match in rules.

    Parameters:
    -----------

    match: Any
        The match to be processed.

    fixed_text: str
        The fixed text.

    cur: int
        The cursor position.

    cur_end: int
        The end cursor position.

    Returns:
    --------

    bool
        True if the match is successful, else False.
    """

    # Initial/default value for replace.
    replace = True

    # Set check cursor depending on match['type']
    chk = cur - 1 if match["type"] == "prefix" else cur_end

    # Set scope based on whether scope is negative.
    if match["scope"].startswith("!"):
        scope = match["scope"][1:]
        negative = True
    else:
        scope = match["scope"]
        negative = False

    # Let the matching begin!
    if scope == "punctuation":
        if (
            not (
                (chk < 0 and match["type"] == "prefix")
                or (chk >= len(fixed_text) and match["type"] == "suffix")
                or validate.is_punctuation(fixed_text[chk])
            )
            != negative
        ):
            replace = False

    elif scope == "vowel":
        if (
            not (
                (
                    (chk >= 0 and match["type"] == "prefix")
                    or (chk < len(fixed_text) and match["type"] == "suffix")
                )
                and validate.is_vowel(fixed_text[chk])
            )
            != negative
        ):
            replace = False

    elif scope == "consonant":
        if (
            not (
                (
                    (chk >= 0 and match["type"] == "prefix")
                    or (chk < len(fixed_text) and match["type"] == "suffix")
                )
                and validate.is_consonant(fixed_text[chk])
            )
            != negative
        ):
            replace = False

    elif scope == "exact":
        if match["type"] == "prefix":
            exact_start = cur - len(match["value"])
            exact_end = cur
        else:
            exact_start = cur_end
            exact_end = cur_end + len(match["value"])

        if not validate.is_exact(
            match["value"], fixed_text, exact_start, exact_end, negative
        ):
            replace = False

    return replace


def rearrange_unicode_text(text: str) -> str:
    """Rearranges Unicode (Avro) text to match conversion standards for ASCII.

    Parameters:
    -----------

    text: str
        The text to be rearranged.

    Returns:
    --------

    str
        The rearranged text.
    """

    # Convert the string to a list of individual characters.
    chars = list(text)
    length = len(chars)
    barrier = 0

    for i in range(length):
        if validate.is_bangla_prekar(chars[i]):
            j = 1

            while (
                i - j >= 0
                and i - j > barrier
                and validate.is_bangla_banjonborno(chars[i - j])
                and validate.is_bangla_halant(chars[i - j - 1])
            ):
                j += 2

            chars[i - j], chars[i] = chars[i], chars[i - j]
            barrier = i + 1

        if (
            i < length - 1
            and validate.is_bangla_halant(chars[i])
            and chars[i - 1] == "র"
            and not validate.is_bangla_halant(chars[i - 2])
        ):
            j = 1
            found_pre_kar = 0

            while True:
                if validate.is_bangla_banjonborno(chars[i + j]):
                    if validate.is_bangla_halant(chars[i + j + 1]):
                        j += 2
                    elif validate.is_bangla_prekar(chars[i + j + 1]):
                        found_pre_kar = 1
                        break
                else:
                    break

            (
                chars[i - 1],
                chars[i],
                chars[i + 1 : i + j + found_pre_kar + 1],
                chars[i + j + 1 :],
            ) = (
                chars[i + j + 1],
                chars[i + 1 : i + j + 1],
                chars[i - 1],
                chars[i],
                chars[i + j + found_pre_kar + 1 :],
            )
            i += j + found_pre_kar
            barrier = i + 1

    return "".join(chars)


def rearrange_bijoy_text(text: str) -> str:
    """Rearranges Bijoy Keyboard text to match conversion standards for Unicode.

    Parameters:
    -----------

    text: str
        The text to be rearranged.

    Returns:
    --------

    str
        The rearranged text.
    """

    i = 0
    while i < len(text):
        if (
            i > 0
            and text[i] == "\u09cd"
            and (
                validate.is_bangla_kar(text[i - 1])
                or validate.is_bangla_nukta(text[i - 1])
            )
            and i < len(text) - 1
        ):
            text = (
                text[: i - 1]
                + text[i]
                + text[i + 1]
                + text[i - 1]
                + text[i + 2 :]
            )

        if (
            0 < i < len(text) - 1
            and text[i] == "\u09cd"
            and text[i - 1] == "\u09b0"
            and text[i - 2] != "\u09cd"
            and validate.is_bangla_kar(text[i + 1])
        ):
            text = (
                text[: i - 1]
                + text[i + 1]
                + text[i - 1]
                + text[i]
                + text[i + 2 :]
            )

        if (
            i < len(text) - 1
            and text[i] == "র"
            and validate.is_bangla_halant(text[i + 1])
            and not validate.is_bangla_halant(text[i - 1])
        ):
            j = 1
            while True:
                if i - j < 0:
                    break
                if validate.is_bangla_banjonborno(
                    text[i - j]
                ) and validate.is_bangla_halant(text[i - j - 1]):
                    j += 2
                elif j == 1 and validate.is_bangla_kar(text[i - j]):
                    j += 1
                else:
                    break
            text = (
                text[: i - j]
                + text[i]
                + text[i + 1]
                + text[i - j : i]
                + text[i + 2 :]
            )
            i += 1
            continue

        if (
            i < len(text) - 1
            and validate.is_bangla_prekar(text[i])
            and text[i + 1] != " "
        ):
            j = 1
            while validate.is_bangla_banjonborno(text[i + j]):
                if validate.is_bangla_halant((part := text[i + j + 1])):
                    j += 2
                else:
                    break
            temp = text[:i] + text[i + 1 : i + j + 1]
            if text[i] == "ে" and part == "া":
                temp += "ো"
                add = 1
            elif text[i] == "ে" and part == "ৗ":
                temp += "ৌ"
                add = 1
            else:
                temp += text[i]
                add = 0
            temp += text[i + j + add + 1 :]
            text = temp
            i += j

        if (
            i < len(text) - 1
            and text[i] == "ঁ"
            and validate.is_bangla_postkar(text[i + 1])
        ):
            temp = text[:i] + text[i + 1] + text[i] + text[i + 2 :]
            text = temp

        i += 1

    return text



# SPDX-License-Identifier: MIT


"""The core module for the avro.py package.

avro.py
Licensed under the terms of the MIT License.
"""

# Imports.
import asyncio
import re
from concurrent.futures import ThreadPoolExecutor
from functools import lru_cache
from itertools import chain
from typing import Callable, Generator, Union

from .core import processor, validate
from .core.config import BIJOY_MAP, BIJOY_MAP_REVERSE

# Compiled regex patterns. These are primarily used in parse() and reverse()
# function calls to validate input text and search for invalid UTF-8 characters.
UTF8_REGEX = re.compile(r"\A[\x00-\x7F]*\Z", re.UNICODE)
REVERSE_REGEX = re.compile(r"(\s|\.|,|\?|।|-|;|')", re.UNICODE)


# This is a backend function and MUST NOT BE EXPORTED!
async def _async_concurrency_helper(
    func: Callable, params: tuple[str, ...]
) -> list[str]:
    """Concurrency helper for the core functions of avro.py.

    Parameters:
    -----------

    func: Callable
        The function to run concurrently.

    params: tuple[str, ...]
        The parameters to pass to the function.

    Returns:
    --------

    list[str]
        The results of the function run concurrently.

    """
    loop = asyncio.get_event_loop()
    tasks = [loop.run_in_executor(None, func, text) for text in params]
    results = await asyncio.gather(*tasks)
    return results


# This is a backend function and MUST NOT BE EXPORTED!
def _sync_concurrency_helper(
    func: Callable, params: tuple[str, ...]
) -> list[str]:
    """Synchronous concurrency helper for the core functions of avro.py using multithreading.

    Parameters:
    -----------

    func: Callable
        The function to run concurrently.

    params: tuple[str, ...]
        The parameters to pass to the function.

    Returns:
    --------

    list[str]
        The results of the function run concurrently.

    """
    with ThreadPoolExecutor() as executor:
        results = list(executor.map(func, params))
    return results


# This is a backend function and MUST NOT BE EXPORTED!
def _parse_output_generator(
    fixed_text: str, cur_end: int
) -> Generator[str, None, None]:
    """Output generator for the parse() function.

    Parameters:
    -----------

    fixed_text: str
        The fixed text to parse.

    cur_end: int
        The cursor end point.

    Yields:
    -------

    str
        The parsed output.
    """

    for cur, i in enumerate(fixed_text):
        uni_pass = UTF8_REGEX.match(i) is not None
        if not uni_pass:
            cur_end = cur + 1
            yield i
        elif cur >= cur_end and uni_pass:
            match = processor.match_patterns(fixed_text, cur, rule=False)
            matched = match["matched"]
            if matched:
                yield match["replaced"]
                cur_end = cur + len(match["found"])
            else:
                match = processor.match_patterns(fixed_text, cur, rule=True)
                matched = match["matched"]
                if matched:
                    cur_end = cur + len(match["found"])
                    replaced = processor.process_rules(
                        rules=match["rules"],
                        fixed_text=fixed_text,
                        cur=cur,
                        cur_end=cur_end,
                    )
                    if replaced:
                        yield replaced
                    else:
                        yield match["replaced"]
            if not matched:
                cur_end = cur + 1
                yield i


# This is a backend function and MUST NOT BE EXPORTED!
@lru_cache(maxsize=128)
def _parse_backend(text: str, remap_words: bool) -> str:
    """The working backend for the parse() function.

    Parameters:
    -----------

    text: str
        The text to parse.

    remap_words: bool
        Whether to parse input text with remapped (exception) words.

    Returns:
    --------

    str
        The parsed text.
    """

    fixed_text = validate.fix_string_case(text)
    manual_required = True
    cur_end = 0

    # Replace predefined exceptions in the input text.
    if remap_words:
        fixed_text, manual_required = processor.find_in_remap(fixed_text)

    return (
        "".join(
            chain.from_iterable(_parse_output_generator(fixed_text, cur_end))
        )
        if manual_required
        else fixed_text
    )


# This is a backend function and MUST NOT BE EXPORTED!
@lru_cache(maxsize=128)
def _convert_backend(text: str) -> str:
    """The working backend for the to_bijoy() function.

    Parameters:
    -----------

    text: str
        The text to convert.

    Returns:
    --------

    str
        The converted text.
    """

    text = processor.rearrange_unicode_text(
        re.sub("ৌ", "ৌ", re.sub("ো", "ো", text))
    )

    for unic in BIJOY_MAP:
        text = re.sub(unic, BIJOY_MAP[unic], text)

    return text.strip()


# This is a backend function and MUST NOT BE EXPORTED!
@lru_cache(maxsize=128)
def _convert_backend_unicode(text: str) -> str:
    """The working backend for the to_unicode() function.

    Parameters:
    -----------

    text: str
        The text to convert.

    Returns:
    --------

    str
        The converted text.
    """

    for ascii_c in BIJOY_MAP_REVERSE:
        text = re.sub(re.escape(ascii_c), BIJOY_MAP_REVERSE[ascii_c], text)

    text = re.sub("অা", "আ", processor.rearrange_bijoy_text(text))
    return text.strip()


# This is a backend function and MUST NOT BE EXPORTED!
def _reverse_output_generator(text: str) -> Generator[str, None, None]:
    """Generates the reversed output for the reverse() function.

    Parameters:
    -----------
    text: str
        The text to reverse.

    Yields:
    -------
    str
        The reversed output.
    """

    for cur, i in enumerate(text):
        try:
            i.encode("utf-8")
            match = processor.match_patterns(
                text, cur, rule=False, reversed=True
            )
            yield (
                (match["reversed"] or match["found"])
                if match["matched"]
                else i
            )
        except UnicodeDecodeError:
            yield i


# This is a backend function and MUST NOT BE EXPORTED!
@lru_cache(maxsize=128)
def _reverse_backend(text: str, remap_words: bool) -> str:
    """The working backend for the reverse() function.

    Parameters:
    -----------
    text: str
        The text to reverse.

    remap_words: bool
        Whether to reverse input text with remapped (exception) words.

    Returns:
    --------
    str
        The reversed text.
    """

    manual_required = True

    if remap_words:
        text, manual_required = processor.find_in_remap(text, reversed=True)

    return (
        "".join(chain.from_iterable(_reverse_output_generator(text)))
        if manual_required
        else text
    )


# This is a backend function and MUST NOT BE EXPORTED!
@lru_cache(maxsize=128)
def _reverse_backend_ext(text: str, remap_words: bool) -> str:
    """Backend extension for the reverse() function.

    Parameters:
    -----------
    text: str
        The text to reverse.

    remap_words: bool
        Whether to reverse input text with remapped (exception) words.

    Returns:
    --------
    str
        The reversed text.
    """

    separated_texts = REVERSE_REGEX.split(text)
    text_segments = [
        _reverse_backend(separated_text, remap_words)
        for separated_text in separated_texts
    ]
    return "".join(text_segments)


# ---

# Primary user-end functions.

# ---


async def parse_async(
    *texts: str, bijoy: bool = False, remap_words: bool = True
) -> Union[str, list[str]]:
    """Asynchronous version of parse() function.

    Parses input text, matches and replaces using the Avro Dictionary.

    If a valid replacement is found, then it returns the replaced string.
    If no replacement is found, then it instead returns the input text.

    Parameters:
    -----------
    *texts: str
        The text(s) to parse.

    bijoy: bool = False
        Whether to return result in the Bijoy Keyboard format (ASCII).

    remap_words: bool = True
        Whether to parse input text with remapped (exception) words.

    Returns:
    --------

    str | list[str]
        The parsed text(s).
    """

    output = await _async_concurrency_helper(
        lambda text: _parse_backend(text, remap_words), texts
    )

    # If the `bijoy` parameter is set to `True`, then convert the output to Bijoy Keyboard format.
    if bijoy:
        return await to_bijoy_async(*output)
    else:
        return output[0] if len(output) == 1 else output


def parse(
    *texts: str, bijoy: bool = False, remap_words: bool = True
) -> Union[str, list[str]]:
    """Parses input text, matches and replaces using the Avro Dictionary.

    If a valid replacement is found, then it returns the replaced string.
    If no replacement is found, then it instead returns the input text.

    Parameters:
    -----------
    *texts: str
        The text(s) to parse.

    bijoy: bool = False
        Whether to return result in the Bijoy Keyboard format (ASCII).

    remap_words: bool = True
        Whether to parse input text with remapped (exception) words.

    Returns:
    --------

    str | list[str]
        The parsed text(s).
    """

    output = _sync_concurrency_helper(
        lambda text: _parse_backend(text, remap_words), texts
    )

    if bijoy:
        return to_bijoy(*output)
    else:
        return output[0] if len(output) == 1 else output


async def to_bijoy_async(*texts: str) -> Union[str, list[str]]:
    """Asynchronous version of to_bijoy() function.

    Converts input text (Avro, Unicode) to Bijoy Keyboard format (ASCII).

    If a valid conversion is found, then it returns the converted string.

    Parameters:
    -----------

    *texts: str
        The text(s) to convert.

    Returns:
    --------

    str | list[str]
        The converted text(s).
    """

    output = await _async_concurrency_helper(_convert_backend, texts)
    return output[0] if len(output) == 1 else output


def to_bijoy(*texts: str) -> Union[str, list[str]]:
    """Converts input text (Avro, Unicode) to Bijoy Keyboard format (ASCII).

    If a valid conversion is found, then it returns the converted string.

    Parameters:
    -----------

    *texts: str
        The text(s) to convert.

    Returns:
    --------

    str | list[str]
        The converted text(s).
    """

    output = _sync_concurrency_helper(_convert_backend, texts)
    return output[0] if len(output) == 1 else output


async def to_unicode_async(*texts: str) -> Union[str, list[str]]:
    """Asynchronous version of to_unicode() function.

    Converts input text (Bijoy Keyboard, ASCII) to Unicode (Avro Keyboard format).

    If a valid conversion is found, then it returns the converted string.

    Parameters:
    -----------
    *texts: str
        The text(s) to convert.

    Returns:
    --------
    str | list[str]
        The converted text(s).
    """

    output = await _async_concurrency_helper(_convert_backend_unicode, texts)
    return output[0] if len(output) == 1 else output


def to_unicode(*texts: str) -> Union[str, list[str]]:
    """Converts input text (Bijoy Keyboard, ASCII) to Unicode (Avro Keyboard format).

    If a valid conversion is found, then it returns the converted string.

    Parameters:
    -----------
    *texts: str
        The text(s) to convert.

    Returns:
    --------
    str | list[str]
        The converted text(s).
    """

    output = _sync_concurrency_helper(_convert_backend_unicode, texts)
    return output[0] if len(output) == 1 else output


async def reverse_async(
    *texts: str, from_bijoy: bool = False, remap_words: bool = True
) -> Union[str, list[str]]:
    """Asynchronous version of reverse() function.

    Reverses input text to Roman script typed in English.

    If a valid replacement is found, then it returns the replaced string.
    If no replacement is found, then it instead returns the input text.

    Parameters:
    -----------

    *texts: str
        The text(s) to reverse.

    from_bijoy: bool = False
        Whether to reverse input text from Bijoy Keyboard format (ASCII).

    remap_words: bool = True
        Whether to reverse input text with remapped (exception) words.

    Returns:
    --------

    str | list[str]
        The reversed text(s).
    """

    # Convert from Bijoy to Unicode if from_bijoy is True
    if from_bijoy:
        converted_texts = await to_unicode_async(*texts)
        if isinstance(converted_texts, str):
            texts = (converted_texts,)

    output = await _async_concurrency_helper(
        lambda text: _reverse_backend_ext(text, remap_words), texts
    )
    return output[0] if len(output) == 1 else output


def reverse(
    *texts: str, from_bijoy: bool = False, remap_words: bool = True
) -> Union[str, list[str]]:
    """Reverses input text to Roman script typed in English.

    If a valid replacement is found, then it returns the replaced string.
    If no replacement is found, then it instead returns the input text.

    Parameters:
    -----------

    *texts: str
        The text(s) to reverse.

    from_bijoy: bool = False
        Whether to reverse input text from Bijoy Keyboard format (ASCII).

    remap_words: bool = True
        Whether to reverse input text with remapped (exception) words.

    Returns:
    --------

    str | list[str]
        The reversed text(s).
    """

    # Convert from Bijoy to Unicode if from_bijoy is True
    if from_bijoy:
        converted_texts = to_unicode(*texts)
        if isinstance(converted_texts, str):
            texts = (converted_texts,)

    output = _sync_concurrency_helper(
        lambda text: _reverse_backend_ext(text, remap_words), texts
    )
    return output[0] if len(output) == 1 else output



# SPDX-License-Identifier: MIT


# Imports.
from typing import Any

# The Avro Dictionary, implemented in Python.
DICT: dict[str, Any] = {
    # Metadata.
    "meta": {
        "file_name": "dictionary.py",
        "file_description": "Provides Avro Dictionary in native Python.",
        "package": "avro.py",
        "license": "MIT License",
        "source": "https://github.com/hitblast/avro.py/blob/main/avro/resources/dictionary.py",
        "adapted_by": "HitBlast",
        "encoding": "utf-8",
    },
    # Avro Keyboard mappings.
    "avro": {
        "patterns": [
            {"find": "bhl", "replace": "ভ্ল"},
            {"find": "psh", "replace": "পশ"},
            {"find": "bdh", "replace": "ব্ধ", "reverse": "bdh"},
            {"find": "bj", "replace": "ব্জ", "reverse": "bj"},
            {"find": "bd", "replace": "ব্দ", "reverse": "bd"},
            {"find": "bb", "replace": "ব্ব", "reverse": "bb"},
            {"find": "bl", "replace": "ব্ল", "reverse": "bl"},
            {"find": "bh", "replace": "ভ", "reverse": "bh"},
            {"find": "vl", "replace": "ভ্ল", "reverse": "vl"},
            {"find": "b", "replace": "ব", "reverse": "b"},
            {"find": "v", "replace": "ভ", "reverse": "bh"},
            {"find": "cNG", "replace": "চ্ঞ", "reverse": "cng"},
            {"find": "cch", "replace": "চ্ছ", "reverse": "cch"},
            {"find": "cc", "replace": "চ্চ", "reverse": "cc"},
            {"find": "ch", "replace": "ছ"},
            {"find": "c", "replace": "চ", "reverse": "ch"},
            {"find": "dhn", "replace": "ধ্ন", "reverse": "dhn"},
            {"find": "dhm", "replace": "ধ্ম", "reverse": "dhm"},
            {"find": "dgh", "replace": "দ্ঘ", "reverse": "dgh"},
            {"find": "ddh", "replace": "দ্ধ", "reverse": "ddh"},
            {"find": "dbh", "replace": "দ্ভ", "reverse": "dv"},
            {"find": "dv", "replace": "দ্ভ", "reverse": "dv"},
            {"find": "dm", "replace": "দ্ম", "reverse": "dd"},
            {"find": "DD", "replace": "ড্ড", "reverse": "dd"},
            {"find": "Dh", "replace": "ঢ", "reverse": "dh"},
            {"find": "dh", "replace": "ধ", "reverse": "dh"},
            {"find": "dg", "replace": "দ্গ"},
            {"find": "dd", "replace": "দ্দ", "reverse": "dd"},
            {"find": "D", "replace": "ড", "reverse": "d"},
            {"find": "d", "replace": "দ", "reverse": "d"},
            {"find": "...", "replace": "..."},
            {"find": ".`", "replace": ".", "reverse": "."},
            {"find": "..", "replace": "।।"},
            {"find": ".", "replace": "।", "reverse": "."},
            {"find": "ghn", "replace": "ঘ্ন", "reverse": "ghn"},
            {"find": "Ghn", "replace": "ঘ্ন", "reverse": "ghn"},
            {"find": "gdh", "replace": "গ্ধ", "reverse": "gdh"},
            {"find": "Gdh", "replace": "গ্ধ", "reverse": "gdh"},
            {"find": "gN", "replace": "গ্ণ", "reverse": "gn"},
            {"find": "GN", "replace": "গ্ণ", "reverse": "gn"},
            {"find": "gn", "replace": "গ্ন", "reverse": "gn"},
            {"find": "Gn", "replace": "গ্ন", "reverse": "gn"},
            {"find": "gm", "replace": "গ্ম"},
            {"find": "Gm", "replace": "গ্ম", "reverse": "gm"},
            {"find": "gl", "replace": "গ্ল", "reverse": "gl"},
            {"find": "Gl", "replace": "গ্ল", "reverse": "gl"},
            {"find": "gg", "replace": "জ্ঞ", "reverse": "gg"},
            {"find": "GG", "replace": "জ্ঞ", "reverse": "gg"},
            {"find": "Gg", "replace": "জ্ঞ", "reverse": "gg"},
            {"find": "gG", "replace": "জ্ঞ", "reverse": "gg"},
            {"find": "gh", "replace": "ঘ", "reverse": "gh"},
            {"find": "Gh", "replace": "ঘ", "reverse": "gh"},
            {"find": "g", "replace": "গ", "reverse": "g"},
            {"find": "G", "replace": "গ", "reverse": "g"},
            {"find": "hN", "replace": "হ্ণ", "reverse": "nn"},
            {"find": "hn", "replace": "হ্ন", "reverse": "nn"},
            {"find": "hm", "replace": "হ্ম", "reverse": "mm"},
            {"find": "hl", "replace": "হ্ল"},
            {"find": "h", "replace": "হ", "reverse": "h"},
            {"find": "jjh", "replace": "জ্ঝ"},
            {"find": "jNG", "replace": "জ্ঞ", "reverse": "gg"},
            {"find": "jh", "replace": "ঝ", "reverse": "jh"},
            {"find": "jj", "replace": "জ্জ", "reverse": "jj"},
            {"find": "j", "replace": "জ", "reverse": "j"},
            {"find": "J", "replace": "জ", "reverse": "j"},
            {"find": "kkhN", "replace": "ক্ষ্ণ", "reverse": "kkhn"},
            {"find": "kShN", "replace": "ক্ষ্ণ", "reverse": "kkhn"},
            {"find": "kkhm", "replace": "ক্ষ্ম"},
            {"find": "kShm", "replace": "ক্ষ্ম", "reverse": "kkh"},
            {"find": "kxN", "replace": "ক্ষ্ণ", "reverse": "kkh"},
            {"find": "kxm", "replace": "ক্ষ্ম", "reverse": "kkh"},
            {"find": "kkh", "replace": "ক্ষ", "reverse": "kkh"},
            {"find": "kSh", "replace": "ক্ষ", "reverse": "kkh"},
            {"find": "ksh", "replace": "কশ"},
            {"find": "kx", "replace": "ক্ষ", "reverse": "kkh"},
            {"find": "kk", "replace": "ক্ক", "reverse": "kk"},
            {"find": "kT", "replace": "ক্ট", "reverse": "kt"},
            {"find": "kt", "replace": "ক্ত", "reverse": "kt"},
            {"find": "kl", "replace": "ক্ল", "reverse": "kl"},
            {"find": "ks", "replace": "ক্স", "reverse": "ks"},
            {"find": "kh", "replace": "খ", "reverse": "kh"},
            {"find": "k", "replace": "ক", "reverse": "k"},
            {"find": "lbh", "replace": "ল্ভ"},
            {"find": "ldh", "replace": "ল্ধ"},
            {"find": "lkh", "replace": "লখ"},
            {"find": "lgh", "replace": "লঘ"},
            {"find": "lph", "replace": "লফ"},
            {"find": "lk", "replace": "ল্ক", "reverse": "lk"},
            {"find": "lg", "replace": "ল্গ"},
            {"find": "lT", "replace": "ল্ট", "reverse": "lt"},
            {"find": "lD", "replace": "ল্ড", "reverse": "ld"},
            {"find": "lp", "replace": "ল্প", "reverse": "lp"},
            {"find": "lv", "replace": "ল্ভ"},
            {"find": "lm", "replace": "ল্ম", "reverse": "lm"},
            {"find": "ll", "replace": "ল্ল", "reverse": "ll"},
            {"find": "lb", "replace": "ল্ব", "reverse": "lb"},
            {"find": "l", "replace": "ল", "reverse": "l"},
            {"find": "mth", "replace": "ম্থ"},
            {"find": "mph", "replace": "ম্ফ", "reverse": "mf"},
            {"find": "mbh", "replace": "ম্ভ", "reverse": "mv"},
            {"find": "mpl", "replace": "মপ্ল"},
            {"find": "mn", "replace": "ম্ন", "reverse": "mn"},
            {"find": "mp", "replace": "ম্প", "reverse": "mp"},
            {"find": "mv", "replace": "ম্ভ", "reverse": "mv"},
            {"find": "mm", "replace": "ম্ম", "reverse": "mm"},
            {"find": "ml", "replace": "ম্ল", "reverse": "ml"},
            {"find": "mb", "replace": "ম্ব", "reverse": "mb"},
            {"find": "mf", "replace": "ম্ফ", "reverse": "mf"},
            {"find": "m", "replace": "ম", "reverse": "m"},
            {"find": "0", "replace": "০", "reverse": "0"},
            {"find": "1", "replace": "১", "reverse": "1"},
            {"find": "2", "replace": "২", "reverse": "2"},
            {"find": "3", "replace": "৩", "reverse": "3"},
            {"find": "4", "replace": "৪", "reverse": "4"},
            {"find": "5", "replace": "৫", "reverse": "5"},
            {"find": "6", "replace": "৬", "reverse": "6"},
            {"find": "7", "replace": "৭", "reverse": "7"},
            {"find": "8", "replace": "৮", "reverse": "8"},
            {"find": "9", "replace": "৯", "reverse": "9"},
            {"find": "NgkSh", "replace": "ঙ্ক্ষ", "reverse": "ngkh"},
            {"find": "Ngkkh", "replace": "ঙ্ক্ষ", "reverse": "ngkh"},
            {"find": "NGch", "replace": "ঞ্ছ", "reverse": "ngch"},
            {"find": "Nggh", "replace": "ঙ্ঘ"},
            {"find": "Ngkh", "replace": "ঙ্খ", "reverse": "ngkh"},
            {"find": "NGjh", "replace": "ঞ্ঝ"},
            {"find": "ngOU", "replace": "ঙ্গৌ"},
            {"find": "ngOI", "replace": "ঙ্গৈ"},
            {"find": "Ngkx", "replace": "ঙ্ক্ষ", "reverse": "ngkh"},
            {"find": "NGc", "replace": "ঞ্চ", "reverse": "nch"},
            {"find": "nch", "replace": "ঞ্ছ", "reverse": "ngch"},
            {"find": "njh", "replace": "ঞ্ঝ"},
            {"find": "ngh", "replace": "ঙ্ঘ"},
            {"find": "Ngk", "replace": "ঙ্ক", "reverse": "ngk"},
            {"find": "Ngx", "replace": "ঙ্ষ"},
            {"find": "Ngg", "replace": "ঙ্গ", "reverse": "ngg"},
            {"find": "Ngm", "replace": "ঙ্ম"},
            {"find": "NGj", "replace": "ঞ্জ", "reverse": "ngj"},
            {"find": "ndh", "replace": "ন্ধ", "reverse": "ndh"},
            {"find": "nTh", "replace": "ন্ঠ", "reverse": "nth"},
            {"find": "NTh", "replace": "ণ্ঠ", "reverse": "nth"},
            {"find": "nth", "replace": "ন্থ", "reverse": "nth"},
            {"find": "nkh", "replace": "ঙ্খ", "reverse": "ngkh"},
            {"find": "ngo", "replace": "ঙ্গ", "reverse": "ngg"},
            {"find": "nga", "replace": "ঙ্গা"},
            {"find": "ngi", "replace": "ঙ্গি"},
            {"find": "ngI", "replace": "ঙ্গী"},
            {"find": "ngu", "replace": "ঙ্গু"},
            {"find": "ngU", "replace": "ঙ্গূ"},
            {"find": "nge", "replace": "ঙ্গে"},
            {"find": "ngO", "replace": "ঙ্গো"},
            {"find": "NDh", "replace": "ণ্ঢ"},
            {"find": "nsh", "replace": "নশ"},
            {"find": "Ngr", "replace": "ঙর"},
            {"find": "NGr", "replace": "ঞর"},
            {"find": "ngr", "replace": "ংর"},
            {"find": "nj", "replace": "ঞ্জ", "reverse": "ngj"},
            {"find": "Ng", "replace": "ঙ", "reverse": "ng"},
            {"find": "NG", "replace": "ঞ", "reverse": "y"},
            {"find": "nk", "replace": "ঙ্ক", "reverse": "ngk"},
            {"find": "ng", "replace": "ং", "reverse": "ng"},
            {"find": "nn", "replace": "ন্ন", "reverse": "nn"},
            {"find": "NN", "replace": "ণ্ণ"},
            {"find": "Nn", "replace": "ণ্ন"},
            {"find": "nm", "replace": "ন্ম", "reverse": "nm"},
            {"find": "Nm", "replace": "ণ্ম"},
            {"find": "nd", "replace": "ন্দ", "reverse": "nd"},
            {"find": "nT", "replace": "ন্ট", "reverse": "nt"},
            {"find": "NT", "replace": "ণ্ট", "reverse": "nt"},
            {"find": "nD", "replace": "ন্ড", "reverse": "nd"},
            {"find": "ND", "replace": "ণ্ড", "reverse": "nd"},
            {"find": "nt", "replace": "ন্ত", "reverse": "nt"},
            {"find": "ns", "replace": "ন্স"},
            {"find": "nc", "replace": "ঞ্চ", "reverse": "nch"},
            {"find": "n", "replace": "ন", "reverse": "n"},
            {"find": "N", "replace": "ণ", "reverse": "n"},
            {"find": "OI`", "replace": "ৈ", "reverse": "oi"},
            {"find": "OU`", "replace": "ৌ", "reverse": "ou"},
            {"find": "O`", "replace": "ো", "reverse": "o"},
            {
                "find": "OI",
                "replace": "ৈ",
                "rules": [
                    {
                        "matches": [{"type": "prefix", "scope": "!consonant"}],
                        "replace": "ঐ",
                    },
                    {
                        "matches": [
                            {"type": "prefix", "scope": "punctuation"}
                        ],
                        "replace": "ঐ",
                    },
                ],
                "reverse": "oi",
            },
            {
                "find": "OU",
                "replace": "ৌ",
                "rules": [
                    {
                        "matches": [{"type": "prefix", "scope": "!consonant"}],
                        "replace": "ঔ",
                    },
                    {
                        "matches": [
                            {"type": "prefix", "scope": "punctuation"}
                        ],
                        "replace": "ঔ",
                    },
                ],
                "reverse": "ou",
            },
            {
                "find": "O",
                "replace": "ো",
                "rules": [
                    {
                        "matches": [{"type": "prefix", "scope": "!consonant"}],
                        "replace": "ও",
                    },
                    {
                        "matches": [
                            {"type": "prefix", "scope": "punctuation"}
                        ],
                        "replace": "ও",
                    },
                ],
                "reverse": "o",
            },
            {"find": "phl", "replace": "ফ্ল", "reverse": "fl"},
            {"find": "pT", "replace": "প্ট", "reverse": "pt"},
            {"find": "pt", "replace": "প্ত", "reverse": "pt"},
            {"find": "pn", "replace": "প্ন", "reverse": "pn"},
            {"find": "pp", "replace": "প্প", "reverse": "pp"},
            {"find": "pl", "replace": "প্ল", "reverse": "pl"},
            {"find": "ps", "replace": "প্স", "reverse": "ps"},
            {"find": "ph", "replace": "ফ", "reverse": "ph"},
            {"find": "fl", "replace": "ফ্ল", "reverse": "fl"},
            {"find": "f", "replace": "ফ", "reverse": "ph"},
            {"find": "p", "replace": "প", "reverse": "p"},
            {"find": "rri`", "replace": "ৃ", "reverse": "ri"},
            {
                "find": "rri",
                "replace": "ৃ",
                "rules": [
                    {
                        "matches": [{"type": "prefix", "scope": "!consonant"}],
                        "replace": "ঋ",
                    },
                    {
                        "matches": [
                            {"type": "prefix", "scope": "punctuation"}
                        ],
                        "replace": "ঋ",
                    },
                ],
                "reverse": "ri",
            },
            {"find": "rrZ", "replace": "রর‍্য"},
            {"find": "rry", "replace": "রর‍্য"},
            {
                "find": "rZ",
                "replace": "র‍্য",
                "rules": [
                    {
                        "matches": [
                            {"type": "prefix", "scope": "consonant"},
                            {
                                "type": "prefix",
                                "scope": "!exact",
                                "value": "r",
                            },
                            {
                                "type": "prefix",
                                "scope": "!exact",
                                "value": "y",
                            },
                            {
                                "type": "prefix",
                                "scope": "!exact",
                                "value": "w",
                            },
                            {
                                "type": "prefix",
                                "scope": "!exact",
                                "value": "x",
                            },
                        ],
                        "replace": "্র্য",
                    }
                ],
            },
            {
                "find": "ry",
                "replace": "র‍্য",
                "rules": [
                    {
                        "matches": [
                            {"type": "prefix", "scope": "consonant"},
                            {
                                "type": "prefix",
                                "scope": "!exact",
                                "value": "r",
                            },
                            {
                                "type": "prefix",
                                "scope": "!exact",
                                "value": "y",
                            },
                            {
                                "type": "prefix",
                                "scope": "!exact",
                                "value": "w",
                            },
                            {
                                "type": "prefix",
                                "scope": "!exact",
                                "value": "x",
                            },
                        ],
                        "replace": "্র্য",
                    }
                ],
            },
            {
                "find": "rr",
                "replace": "রর",
                "rules": [
                    {
                        "matches": [
                            {"type": "prefix", "scope": "!consonant"},
                            {"type": "suffix", "scope": "!vowel"},
                            {
                                "type": "suffix",
                                "scope": "!exact",
                                "value": "r",
                            },
                            {"type": "suffix", "scope": "!punctuation"},
                        ],
                        "replace": "র্",
                    },
                    {
                        "matches": [
                            {"type": "prefix", "scope": "consonant"},
                            {
                                "type": "prefix",
                                "scope": "!exact",
                                "value": "r",
                            },
                        ],
                        "replace": "্রর",
                    },
                ],
            },
            {"find": "Rg", "replace": "ড়্গ"},
            {"find": "Rh", "replace": "ঢ়"},
            {"find": "R", "replace": "ড়", "reverse": "r"},
            {
                "find": "r",
                "replace": "র",
                "rules": [
                    {
                        "matches": [
                            {"type": "prefix", "scope": "consonant"},
                            {
                                "type": "prefix",
                                "scope": "!exact",
                                "value": "r",
                            },
                            {
                                "type": "prefix",
                                "scope": "!exact",
                                "value": "y",
                            },
                            {
                                "type": "prefix",
                                "scope": "!exact",
                                "value": "w",
                            },
                            {
                                "type": "prefix",
                                "scope": "!exact",
                                "value": "x",
                            },
                            {
                                "type": "prefix",
                                "scope": "!exact",
                                "value": "Z",
                            },
                        ],
                        "replace": "্র",
                    }
                ],
            },
            {"find": "shch", "replace": "শ্ছ", "reverse": "sch"},
            {"find": "ShTh", "replace": "ষ্ঠ", "reverse": "sth"},
            {"find": "Shph", "replace": "ষ্ফ", "reverse": "sf"},
            {"find": "Sch", "replace": "শ্ছ", "reverse": "sch"},
            {"find": "skl", "replace": "স্ক্ল"},
            {"find": "skh", "replace": "স্খ"},
            {"find": "sth", "replace": "স্থ", "reverse": "sth"},
            {"find": "sph", "replace": "স্ফ", "reverse": "sf"},
            {"find": "shc", "replace": "শ্চ", "reverse": "scch"},
            {"find": "sht", "replace": "শ্ত"},
            {"find": "shn", "replace": "শ্ন", "reverse": "sn"},
            {"find": "shm", "replace": "শ্ম", "reverse": "ss"},
            {"find": "shl", "replace": "শ্ল", "reverse": "sl"},
            {"find": "Shk", "replace": "ষ্ক", "reverse": "sk"},
            {"find": "ShT", "replace": "ষ্ট", "reverse": "st"},
            {"find": "ShN", "replace": "ষ্ণ", "reverse": "sn"},
            {"find": "Shp", "replace": "ষ্প", "reverse": "sp"},
            {"find": "Shf", "replace": "ষ্ফ", "reverse": "sf"},
            {"find": "Shm", "replace": "ষ্ম", "reverse": "sm"},
            {"find": "spl", "replace": "স্প্ল"},
            {"find": "sk", "replace": "স্ক", "reverse": "sk"},
            {"find": "Sc", "replace": "শ্চ", "reverse": "scch"},
            {"find": "sT", "replace": "স্ট", "reverse": "st"},
            {"find": "st", "replace": "স্ত", "reverse": "st"},
            {"find": "sn", "replace": "স্ন", "reverse": "sn"},
            {"find": "sp", "replace": "স্প", "reverse": "sp"},
            {"find": "sf", "replace": "স্ফ", "reverse": "sf"},
            {"find": "sm", "replace": "স্ম", "reverse": "sh"},
            {"find": "sl", "replace": "স্ল", "reverse": "sl"},
            {"find": "sh", "replace": "শ", "reverse": "sh"},
            {"find": "Sc", "replace": "শ্চ", "reverse": "scch"},
            {"find": "St", "replace": "শ্ত"},
            {"find": "Sn", "replace": "শ্ন", "reverse": "sn"},
            {"find": "Sm", "replace": "শ্ম", "reverse": "ss"},
            {"find": "Sl", "replace": "শ্ল", "reverse": "sl"},
            {"find": "Sh", "replace": "ষ", "reverse": "sh"},
            {"find": "s", "replace": "স", "reverse": "s"},
            {"find": "S", "replace": "শ", "reverse": "sh"},
            {"find": "oo`", "replace": "ু", "reverse": "u"},
            {
                "find": "oo",
                "replace": "ু",
                "rules": [
                    {
                        "matches": [
                            {"type": "prefix", "scope": "!consonant"},
                            {
                                "type": "suffix",
                                "scope": "!exact",
                                "value": "`",
                            },
                        ],
                        "replace": "উ",
                    },
                    {
                        "matches": [
                            {"type": "prefix", "scope": "punctuation"},
                            {
                                "type": "suffix",
                                "scope": "!exact",
                                "value": "`",
                            },
                        ],
                        "replace": "উ",
                    },
                ],
                "reverse": "u",
            },
            # {
            #     "find": "o`"
            #     "replace": ""
            # },
            {"find": "oZ", "replace": "অ্য"},
            {
                "find": "o",
                "replace": "",
                "rules": [
                    {
                        "matches": [
                            {"type": "prefix", "scope": "vowel"},
                            {
                                "type": "prefix",
                                "scope": "!exact",
                                "value": "o",
                            },
                        ],
                        "replace": "ও",
                    },
                    {
                        "matches": [
                            {"type": "prefix", "scope": "vowel"},
                            {"type": "prefix", "scope": "exact", "value": "o"},
                        ],
                        "replace": "অ",
                    },
                    {
                        "matches": [
                            {"type": "prefix", "scope": "punctuation"}
                        ],
                        "replace": "অ",
                    },
                ],
            },
            {"find": "tth", "replace": "ত্থ"},
            {"find": "t``", "replace": "ৎ", "reverse": "t"},
            {"find": "TT", "replace": "ট্ট", "reverse": "tt"},
            {"find": "Tm", "replace": "ট্ম"},
            {"find": "Th", "replace": "ঠ", "reverse": "th"},
            {"find": "tn", "replace": "ত্ন"},
            {"find": "tm", "replace": "ত্ম", "reverse": "tt"},
            {"find": "th", "replace": "থ", "reverse": "th"},
            {"find": "tt", "replace": "ত্ত", "reverse": "tt"},
            {"find": "T", "replace": "ট", "reverse": "t"},
            {"find": "t", "replace": "ত", "reverse": "t"},
            {"find": "aZ", "replace": "অ্যা"},
            {"find": "AZ", "replace": "অ্যা"},
            {"find": "a`", "replace": "া", "reverse": "a"},
            {"find": "A`", "replace": "া", "reverse": "a"},
            {"replace": "আ", "reverse": "a"},
            {"replace": "র", "reverse": "r"},
            {
                "find": "a",
                "replace": "া",
                "rules": [
                    {
                        "matches": [
                            {"type": "prefix", "scope": "punctuation"},
                            {
                                "type": "suffix",
                                "scope": "!exact",
                                "value": "`",
                            },
                        ],
                        "replace": "আ",
                    },
                    {
                        "matches": [
                            {"type": "prefix", "scope": "!consonant"},
                            {
                                "type": "prefix",
                                "scope": "!exact",
                                "value": "a",
                            },
                            {
                                "type": "suffix",
                                "scope": "!exact",
                                "value": "`",
                            },
                        ],
                        "replace": "য়া",
                    },
                    {
                        "matches": [
                            {"type": "prefix", "scope": "exact", "value": "a"},
                            {
                                "type": "suffix",
                                "scope": "!exact",
                                "value": "`",
                            },
                        ],
                        "replace": "আ",
                    },
                ],
                "reverse": "a",
            },
            {"find": "i`", "replace": "ি", "reverse": "i"},
            {"replace": "ই", "reverse": "i"},
            {
                "find": "i",
                "replace": "ি",
                "rules": [
                    {
                        "matches": [
                            {"type": "prefix", "scope": "!consonant"},
                            {
                                "type": "suffix",
                                "scope": "!exact",
                                "value": "`",
                            },
                        ],
                        "replace": "ই",
                    },
                    {
                        "matches": [
                            {"type": "prefix", "scope": "punctuation"},
                            {
                                "type": "suffix",
                                "scope": "!exact",
                                "value": "`",
                            },
                        ],
                        "replace": "ই",
                    },
                ],
            },
            {"find": "I`", "replace": "ী", "reverse": "i"},
            {
                "find": "I",
                "replace": "ী",
                "rules": [
                    {
                        "matches": [
                            {"type": "prefix", "scope": "!consonant"},
                            {
                                "type": "suffix",
                                "scope": "!exact",
                                "value": "`",
                            },
                        ],
                        "replace": "ঈ",
                    },
                    {
                        "matches": [
                            {"type": "prefix", "scope": "punctuation"},
                            {
                                "type": "suffix",
                                "scope": "!exact",
                                "value": "`",
                            },
                        ],
                        "replace": "ঈ",
                    },
                ],
                "reverse": "i",
            },
            {"find": "u`", "replace": "ু", "reverse": "u"},
            {"replace": "উ", "reverse": "u"},
            {
                "find": "u",
                "replace": "ু",
                "rules": [
                    {
                        "matches": [
                            {"type": "prefix", "scope": "!consonant"},
                            {
                                "type": "suffix",
                                "scope": "!exact",
                                "value": "`",
                            },
                        ],
                        "replace": "উ",
                    },
                    {
                        "matches": [
                            {"type": "prefix", "scope": "punctuation"},
                            {
                                "type": "suffix",
                                "scope": "!exact",
                                "value": "`",
                            },
                        ],
                        "replace": "উ",
                    },
                ],
                "reverse": "u",
            },
            {"find": "U`", "replace": "ূ", "reverse": "u"},
            {
                "find": "U",
                "replace": "ূ",
                "rules": [
                    {
                        "matches": [
                            {"type": "prefix", "scope": "!consonant"},
                            {
                                "type": "suffix",
                                "scope": "!exact",
                                "value": "`",
                            },
                        ],
                        "replace": "ঊ",
                    },
                    {
                        "matches": [
                            {"type": "prefix", "scope": "punctuation"},
                            {
                                "type": "suffix",
                                "scope": "!exact",
                                "value": "`",
                            },
                        ],
                        "replace": "ঊ",
                    },
                ],
                "reverse": "u",
            },
            {"find": "ee`", "replace": "ী", "reverse": "i"},
            {
                "find": "ee",
                "replace": "ী",
                "rules": [
                    {
                        "matches": [
                            {"type": "prefix", "scope": "!consonant"},
                            {
                                "type": "suffix",
                                "scope": "!exact",
                                "value": "`",
                            },
                        ],
                        "replace": "ঈ",
                    },
                    {
                        "matches": [
                            {"type": "prefix", "scope": "punctuation"},
                            {
                                "type": "suffix",
                                "scope": "!exact",
                                "value": "`",
                            },
                        ],
                        "replace": "ঈ",
                    },
                ],
                "reverse": "i",
            },
            {"find": "e`", "replace": "ে", "reverse": "e"},
            {"replace": "এ", "reverse": "e"},
            {"replace": "ও", "reverse": "w"},
            {
                "find": "e",
                "replace": "ে",
                "rules": [
                    {
                        "matches": [
                            {"type": "prefix", "scope": "!consonant"},
                            {
                                "type": "suffix",
                                "scope": "!exact",
                                "value": "`",
                            },
                        ],
                        "replace": "এ",
                    },
                    {
                        "matches": [
                            {"type": "prefix", "scope": "punctuation"},
                            {
                                "type": "suffix",
                                "scope": "!exact",
                                "value": "`",
                            },
                        ],
                        "replace": "এ",
                    },
                ],
                "reverse": "e",
            },
            {"find": "z", "replace": "য", "reverse": "z"},
            {"find": "Z", "replace": "্য"},
            {
                "find": "y",
                "replace": "্য",
                "rules": [
                    {
                        "matches": [
                            {"type": "prefix", "scope": "!consonant"},
                            {"type": "prefix", "scope": "!punctuation"},
                        ],
                        "replace": "য়",
                    },
                    {
                        "matches": [
                            {"type": "prefix", "scope": "punctuation"}
                        ],
                        "replace": "ইয়",
                    },
                ],
            },
            {"find": "Y", "replace": "য়", "reverse": "y"},
            {"find": "q", "replace": "ক", "reverse": "k"},
            {
                "find": "w",
                "replace": "ও",
                "rules": [
                    {
                        "matches": [
                            {"type": "prefix", "scope": "punctuation"},
                            {"type": "suffix", "scope": "vowel"},
                        ],
                        "replace": "ওয়",
                    },
                    {
                        "matches": [{"type": "prefix", "scope": "consonant"}],
                        "replace": "্ব",
                    },
                ],
                "reverse": "o",
            },
            {
                "find": "x",
                "replace": "ক্স",
                "rules": [
                    {
                        "matches": [
                            {"type": "prefix", "scope": "punctuation"}
                        ],
                        "replace": "এক্স",
                    }
                ],
                "reverse": "ks",
            },
            {"find": ":`", "replace": ":"},
            {"find": ":", "replace": "ঃ"},
            {"find": "^`", "replace": "^"},
            {"find": "^", "replace": "ঁ", "reverse": ""},
            {"find": ",,", "replace": "্‌"},
            {"find": ",", "replace": ","},
            {"find": "$", "replace": "৳"},
            # {
            #     "find": "`",
            #     "replace": ""
            # }
        ],
        # Remapped words.
        "exceptions": {
            "বাংলাদেশ": "Bangladesh",
            "ইউএসএ": "USA",
            "ইউরোপ": "Europe",
            "আস্ট্রেলিয়া": "Australia",
            "ফেসবুক": "Facebook",
            "গুগল": "Google",
            "উইকিপিডিয়া": "Wikipedia",
            "হোয়াটসঅ্যাপ": "Whatsapp",
            "টুইটার": "Twitter",
            "লিঙ্কডইন": "Linkedin",
            "ইনস্টাগ্রাম": "Instagram",
            "ইউটিউব": "YouTube",
            "আইএমডিবি": "IMDb",
            "অ্যামাজন": "Amazon",
            "মাইক্রোসফট": "Microsoft",
            "অ্যাপল": "Apple",
            "ইউনিভার্সিটি": "University",
            "কম্পিউটার": "Computer",
            "কোম্পানি": "Company",
            "কর্পোরেশন": "Corporation",
            "করোনাভাইরাস": "Coronavirus",
            "আইবিএম": "IBM",
            "ইমেইল": "Email",
            "এসএমএস": "SMS",
        },
        # Constant values.
        "vowel": "aeiou",
        "consonant": "bcdfghjklmnpqrstvwxyz",
        "casesensitive": "oiudgjnrstyz",
        "number": "0123456789",
        # For reverse parsing.
        "shorborno": "অআইঈউঊএঐওঔ",
        "kar": ["া", "ি", "ী", "ৗ", "ু", "ূ", "ৃ", "ে", "ৈ", "ো", "ৌ"],
        # Ignored symbols.
        "ignore": ["ঁ", "।", "?", ".", "-", ";"],
    },
    # Bijoy Keyboard mappings.
    "bijoy": {
        "mappings": {
            "।": "|",
            "‘": "Ô",
            "’": "Õ",
            "“": "Ò",
            "”": "Ó",
            "্র্য": "ª¨",
            "র‌্য": "i¨",
            "ক্ক": "°",
            "ক্ট": "±",
            "ক্ত": "³",
            "ক্ব": "K¡",
            "স্ক্র": "¯Œ",
            "ক্র": "µ",
            "ক্ল": "K¬",
            "ক্ষ": "¶",
            "ক্স": "·",
            "গু": "¸",
            "গ্ধ": "»",
            "গ্ন": "Mœ",
            "গ্ম": "M¥",
            "গ্ল": "M­",
            "গ্রু": "Mªy",
            "ঙ্ক": "¼",
            "ঙ্ক্ষ": "•¶",
            "ঙ্খ": "•L",
            "ঙ্গ": "½",
            "ঙ্ঘ": "•N",
            "চ্চ": "”P",
            "চ্ছ": "”Q",
            "চ্ছ্ব": "”Q¡",
            "চ্ঞ": "”T",
            "জ্জ্ব": "¾¡",
            "জ্জ": "¾",
            "জ্ঝ": "À",
            "জ্ঞ": "Á",
            "জ্ব": "R¡",
            "ঞ্চ": "Â",
            "ঞ্ছ": "Ã",
            "ঞ্জ": "Ä",
            "ঞ্ঝ": "Å",
            "ট্ট": "Æ",
            "ট্ব": "U¡",
            "ট্ম": "U¥",
            "ড্ড": "Ç",
            "ণ্ট": "È",
            "ণ্ঠ": "É",
            "ন্স": "Ý",
            "ণ্ড": "Ê",
            "ন্তু": "š‘",
            "ণ্ব": "Y^",
            "ত্ত": "Ë",
            "ত্ত্ব": "Ë¡",
            "ত্থ": "Ì",
            "ত্ন": "Zœ",
            "ত্ম": "Z¥",
            "ন্ত্ব": "š—¡",
            "ত্ব": "Z¡",
            "থ্ব": "_¡",
            "দ্গ": "˜M",
            "দ্ঘ": "˜N",
            "দ্দ": "Ï",
            "দ্ধ": "×",
            "দ্ব": "˜¡",
            "দ্ব": "Ø",
            "দ্ভ": "™¢",
            "দ্ম": "Ù",
            "দ্রু": "`ª“",
            "ধ্ব": "aŸ",
            "ধ্ম": "a¥",
            "ন্ট": "›U",
            "ন্ঠ": "Ú",
            "ন্ড": "Û",
            "ন্ত্র": "š¿",
            "ন্ত": "š—",
            "স্ত্র": "¯¿",
            "ত্র": "Î",
            "ন্থ": "š’",
            "ন্দ": "›`",
            "ন্দ্ব": "›Ø",
            "ন্ধ": "Ü",
            "ন্ন": "bœ",
            "ন্ব": "š^",
            "ন্ম": "b¥",
            "প্ট": "Þ",
            "প্ত": "ß",
            "প্ন": "cœ",
            "প্প": "à",
            "প্ল": "c­",
            "প্স": "á",
            "ফ্ল": "d¬",
            "ব্জ": "â",
            "ব্দ": "ã",
            "ব্ধ": "ä",
            "ব্ব": "eŸ",
            "ব্ল": "e­",
            "ভ্র": "å",
            "ম্ন": "gœ",
            "ম্প": "¤ú",
            "ম্ফ": "ç",
            "ম্ব": "¤^",
            "ম্ভ": "¤¢",
            "ম্ভ্র": "¤£",
            "ম্ম": "¤§",
            "ম্ল": "¤­",
            "রু": "i“",
            "রূ": "iƒ",
            "ল্ক": "é",
            "ল্গ": "ê",
            "ল্প": "í",
            "ল্ট": "ë",
            "ল্ড": "ì",
            "ল্ফ": "î",
            "ল্ব": "j¦",
            "ল্ম": "j¥",
            "ল্ল": "jø",
            "শু": "ï",
            "শ্চ": "ð",
            "শ্ন": "kœ",
            "শ্ব": "k¦",
            "শ্ম": "k¥",
            "শ্ল": "kø",
            "ষ্ক": "®‹",
            "ষ্ক্র": "®Œ",
            "ষ্ট": "ó",
            "ষ্ঠ": "ô",
            "ষ্ণ": "ò",
            "ষ্প": "®ú",
            "ষ্ফ": "õ",
            "ষ্ম": "®§",
            "স্ক": "¯‹",
            "স্ট": "÷",
            "স্খ": "ö",
            "স্ত": "¯—",
            "স্তু": "¯‘",
            "স্থ": "¯’",
            "স্ন": "mœ",
            "স্প": "¯ú",
            "স্ফ": "ù",
            "স্ব": "¯^",
            "স্ম": "¯§",
            "স্ল": "¯­",
            "হু": "û",
            "হ্ণ": "nè",
            "হ্ন": "ý",
            "হ্ম": "þ",
            "হ্ল": "n¬",
            "হৃ": "ü",
            "র্": "©",
            "্র": "«",
            "্য": "¨",
            "্": "&",
            "আ": "Av",
            "অ": "A",
            "ই": "B",
            "ঈ": "C",
            "উ": "D",
            "ঊ": "E",
            "ঋ": "F",
            "এ": "G",
            "ঐ": "H",
            "ও": "I",
            "ঔ": "J",
            "ক": "K",
            "খ": "L",
            "গ": "M",
            "ঘ": "N",
            "ঙ": "O",
            "চ": "P",
            "ছ": "Q",
            "জ": "R",
            "ঝ": "S",
            "ঞ": "T",
            "ট": "U",
            "ঠ": "V",
            "ড": "W",
            "ঢ": "X",
            "ণ": "Y",
            "ত": "Z",
            "থ": "_",
            "দ": "`",
            "ধ": "a",
            "ন": "b",
            "প": "c",
            "ফ": "d",
            "ব": "e",
            "ভ": "f",
            "ম": "g",
            "য": "h",
            "র": "i",
            "ল": "j",
            "শ": "k",
            "ষ": "l",
            "স": "m",
            "হ": "n",
            "ড়": "o",
            "ঢ়": "p",
            "য়": "q",
            "ৎ": "r",
            "০": "0",
            "১": "1",
            "২": "2",
            "৩": "3",
            "৪": "4",
            "৫": "5",
            "৬": "6",
            "৭": "7",
            "৮": "8",
            "৯": "9",
            "া": "v",
            "ি": "w",
            "ী": "x",
            "ু": "y",
            "ূ": "~",
            "ৃ": "…",
            "ে": "‡",
            "ৈ": "‰",
            "ৗ": "Š",
            "ং": "s",
            "ঃ": "t",
            "ঁ": "u",
        },
        "prekar": ["ি", "ৈ", "ে"],
        "postkar": ["া", "ো", "ৌ", "ৗ", "ু", "ূ", "ী", "ৃ"],
        "banjonborno": "কখগঘঙচছজঝঞটঠডঢণতথদধনপফবভমশষসহযরলয়ংঃঁৎ",
        "exceptions": {
            "halant": "্",
            "nukta": ["ং", "ঃ", "ঁ"],
        },
    },
}
